# -*- coding: utf-8 -*-
"""
lories.data.databases
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from typing import Any, Collection, Optional

import tzlocal

import pandas as pd
from lories.connectors import ConnectorContext, ConnectType, Database
from lories.core import Configurations, Configurator, ResourceError
from lories.data.channels import Channel, Channels
from lories.data.context import DataContext
from lories.data.replication import Replications
from lories.data.retention import Retention, Retentions
from lories.util import floor_date, parse_freq, to_bool, to_timedelta, to_timezone


class Databases(ConnectorContext, Configurator):
    TYPE: str = "databases"

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def __init__(self, context: DataContext, configs: Configurations) -> None:
        super().__init__(context, configs=configs.get_member(Databases.TYPE, defaults={}))
        self.load(configure=False, sort=False)
        self.configure()

        for database in context.connectors.filter(lambda c: c.is_enabled() and isinstance(c, Database)):
            self._add(database)
        self.sort()

    # noinspection PyProtectedMember
    @classmethod
    def _assert_configs(cls, configs: Configurations) -> Configurations:
        if configs is None:
            raise ResourceError(f"Invalid '{cls.__name__}' NoneType configurations")
        return super()._assert_configs(configs)

    def load(self, **kwargs: Any) -> Collection[Database]:
        databases = self._load(
            self,
            self.configs,
            configs_file=self.configs.name,
            configs_dir=self.configs.dirs.conf.joinpath(self.configs.name.replace(".conf", ".d")),
            includes=Database.INCLUDES,
            **kwargs,
        )
        # for database in databases:
        #     database._connect_type = ConnectType.NONE
        return databases

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def connect(self, channels: Optional[Channels] = None) -> None:
        self.context._connect(*self.filter(self.__is_connectable), channels=channels, force=True)

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def _connect(self, database: Database, channels: Optional[Channels] = None) -> None:
        if self.__is_connectable(database):
            self.context._connect(database, channels=channels, force=True)

    # noinspection PyProtectedMember
    @staticmethod
    def __is_connectable(database: Database) -> bool:
        return database._connect_type == ConnectType.NONE or not database._is_connected()

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def disconnect(self):
        self.context._disconnect(*self.filter(self.__is_disconnectable))

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def _disconnect(self, database: Database) -> None:
        if self.__is_disconnectable(database):
            self.context._disconnect(database)

    # noinspection PyProtectedMember
    @staticmethod
    def __is_disconnectable(database: Database) -> bool:
        return database._connect_type == ConnectType.NONE and database._is_connected()

    # noinspection PyProtectedMember
    def replicate(self, channels: Channels, full: bool = False, force: bool = False, **kwargs) -> None:
        replications = Replications()

        def build_replication(channel: Channel) -> Channel:
            channel = channel.from_logger()
            channel.replication = replications.build(self, channel, **kwargs)
            return channel

        channels = channels.apply(build_replication).filter(lambda c: c.replication is not None)

        for database in self.values():
            database_channels = channels.filter(lambda c: c.replication.database.id == database.id)
            if len(database_channels) == 0:
                continue

            self._connect(database, channels=database_channels)
            try:
                for logger, logger_channels in database_channels.groupby(lambda c: c.logger._connector):
                    self._connect(logger, channels=logger_channels)
                    try:
                        for replication, replication_channels in logger_channels.groupby(lambda c: c.replication):
                            replication.replicate(replication_channels, full=to_bool(full), force=to_bool(force))

                    except ResourceError as e:
                        self._logger.warning(f"Error replicating database '{database.id}': {str(e)}")
                        if self._logger.getEffectiveLevel() <= logging.DEBUG:
                            self._logger.exception(e)
                    finally:
                        self._disconnect(logger)
            finally:
                self._disconnect(database)

    # noinspection PyProtectedMember
    def rotate(self, channels: Channels, full: bool = False) -> None:
        retentions = Retentions()

        def build_rotation(channel: Channel) -> Channel:
            channel = channel.from_logger()
            channel.rotate = parse_freq(channel.get("rotate", default=None))
            channel.retentions = Retention.build(self.configs, channel)
            retentions.extend(channel.retentions, unique=True)
            return channel

        channels = channels.apply(build_rotation).filter(lambda c: c.rotate is not None or len(c.retentions) > 0)
        for database in self.values():
            database_channels = channels.filter(lambda c: c.has_logger(database.id))
            if len(database_channels) == 0:
                continue

            self._connect(database, channels=database_channels)
            try:
                for rotation, rotation_channels in database_channels.groupby(lambda c: c.rotate):
                    if rotation is None:
                        continue
                    freq = self.configs.get("freq", default="D")
                    timezone = to_timezone(self.configs.get("timezone", default=tzlocal.get_localzone_name()))
                    rotate = floor_date(pd.Timestamp.now(tz=timezone) - to_timedelta(rotation), freq=freq)

                    for _, deletion_channels in rotation_channels.groupby(lambda c: c.group):
                        start = database.read_first_index(deletion_channels)
                        if start is None or start > rotate:
                            self._logger.debug(
                                f"Skip rotating values of resource{'s' if len(deletion_channels) > 1 else ''} "
                                + ", ".join([f"'{r.id}'" for r in deletion_channels])
                                + " without any values found"
                            )
                            continue

                        self._logger.info(
                            f"Deleting values of resource{'s' if len(deletion_channels) > 1 else ''} "
                            + ", ".join([f"'{r.id}'" for r in deletion_channels])
                            + f" up to {rotate.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        )
                        database.delete(deletion_channels, end=rotate)

                retentions.sort()
                for retention in retentions:
                    if not retention.enabled:
                        self._logger.debug(f"Skipping disabled retention '{retention.keep}'")
                        continue
                    try:
                        # noinspection PyProtectedMember
                        def has_retention(channel: Channel) -> bool:
                            return (
                                channel.logger.enabled
                                and isinstance(channel.logger._connector, Database)
                                and retention in channel.retentions
                            )

                        retention.aggregate(database_channels.filter(has_retention), full=to_bool(full))

                    except ResourceError as e:
                        self._logger.warning(
                            f"Error aggregating '{retention.method}' retaining {retention.keep}: {str(e)}"
                        )
            finally:
                self._disconnect(database)



# -*- coding: utf-8 -*-
"""
lories.io.excel
~~~~~~~~~~~~~~~


"""

import logging
import os
import warnings
from typing import Optional

from openpyxl.styles import Border, Font, Side
from openpyxl.utils import get_column_letter
from tables import NaturalNameWarning

import pandas as pd

warnings.filterwarnings("ignore", category=NaturalNameWarning)
logger = logging.getLogger(__name__)


# noinspection PyShadowingBuiltins
def write(
    path: str,
    sheet_name: str,
    data: pd.DataFrame,
    index: Optional[bool] = None,
) -> None:
    if data.dropna(axis="index", how="all").dropna(axis="columns", how="all").empty:
        return None
    if isinstance(data.index, pd.DatetimeIndex):
        data.index = data.index.tz_localize(None)
    if data.index.name is not None:
        data.index.name = data.index.name.title()
    if index is None:
        index = len(data.index) > 1

    header_length = data.columns.nlevels

    border_side = Side(border_style=None)
    border = Border(
        left=border_side,
        right=border_side,
        bottom=border_side,
        top=border_side,
    )
    font = Font(name="Calibri Light", size=12, color="333333")

    if not path.endswith(".xlsx"):
        path += ".xlsx"

    dir = os.path.dirname(path)
    if not os.path.isdir(dir):
        os.makedirs(dir, exist_ok=True)

    kwargs = {
        "engine": "openpyxl",
    }
    if not os.path.exists(path):
        kwargs["mode"] = "w"
    else:
        kwargs["mode"] = "r+"
        kwargs["if_sheet_exists"] = "replace"

    with pd.ExcelWriter(path, **kwargs) as writer:
        data.to_excel(writer, sheet_name=sheet_name, float_format="%.2f")  # , encoding=encoding)
        data_book = writer.book

        data_sheet = data_book[sheet_name]
        if header_length > 1:
            data_sheet[2][0].value = data_sheet[3][0].value
            data_sheet.delete_rows(3, 1)

        if not index:
            data_sheet.delete_cols(1)
            for data_merged in data_sheet.merged_cells:
                data_merged.shift(col_shift=-1)

        for column in range(1, len(data_sheet[header_length]) + 1):
            column_letter = get_column_letter(column)
            if column == 1 and index:
                column_length = len(data_sheet[column_letter]) + 1
                column_widths = [len(str(i)) for i in data.index.values]
            else:
                column_length = header_length
                column_widths = [len(str(i)) for i in data.iloc[:, column - int(index) - 1].values]

            for column_cell, *_ in data_sheet.iter_rows(
                min_row=1,
                max_row=column_length,
                min_col=column,
                max_col=column,
            ):
                column_cell.border = border
                column_cell.font = font

                if column_cell.row <= header_length:
                    header_cell_str = str(column_cell.value)
                    for header_cell_line in header_cell_str.split("\n"):
                        header_column_width = len(header_cell_line)
                        if header_cell_str in data.columns:
                            header_column_width /= len(data[[header_cell_str]].columns)
                        column_widths.append(header_column_width)

                    if "\n" in str(column_cell.value):
                        # header_alignment = copy(column_cell.alignment)
                        # header_alignment.wrapText = True
                        # header_alignment.vertical = 'center'
                        # header_alignment.horizontal = 'center'
                        # column_cell.alignment = header_alignment
                        column_cell.alignment.wrapText = True
                        data_sheet.row_dimensions[column_cell.row].height = 33

            data_sheet.column_dimensions[column_letter].width = max(column_widths) + 2



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.table
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional, Tuple

import sqlalchemy as sql
from sqlalchemy import ClauseElement, Dialect, Result, UnaryExpression
from sqlalchemy.sql import Delete, Insert, Select, and_, asc, between, desc, func, literal, not_, or_, text
from sqlalchemy.types import BLOB, DATETIME, TIMESTAMP

import numpy as np
import pandas as pd
import pytz as tz
from lories.connectors.sql.columns import Column, DatetimeColumn, SurrogateKeyColumn
from lories.connectors.sql.index import DatetimeIndexType
from lories.core import ResourceError, Resources
from lories.typing import Resource, Timestamp

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class Table(sql.Table):
    def __init__(
        self,
        name: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(name, *args, **kwargs)
        self.datetime_index_type = DatetimeIndexType.NONE
        for datetime_index_type in [
            DatetimeIndexType.DATE_AND_TIME,
            DatetimeIndexType.DATETIME,
            DatetimeIndexType.TIMESTAMP,
            DatetimeIndexType.TIMESTAMP_UNIX,
        ]:
            if self.__has_datetime_index(datetime_index_type):
                self.datetime_index_type = datetime_index_type
                break

    # noinspection PyUnresolvedReferences
    @property
    def dialect(self) -> Dialect:
        return self.metadata.dialect

    @property
    def primary_index(self) -> Column:
        return self.primary_key.columns[0]

    def _primary_order(self, order: Literal["asc", "desc"] = "asc") -> List[UnaryExpression]:
        clauses = []
        for primary_key in reversed(self.primary_key.columns):
            # if isinstance(primary_key, SurrogateKeyColumn):
            #     continue
            if order == "asc":
                clauses.append(asc(primary_key))
            elif order == "desc":
                clauses.append(desc(primary_key))
            else:
                raise ValueError(f"Unknown order '{order}'")
        return clauses

    def _primary_clauses(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> List[ClauseElement]:
        clauses = []
        primary_index = self.primary_index
        if self.__is_datetime_index(primary_index):
            if start is not None and end is not None:
                clauses.append(between(primary_index, primary_index.validate(start), primary_index.validate(end)))
            elif start is not None:
                clauses.append(primary_index >= primary_index.validate(start))
            elif end is not None:
                clauses.append(primary_index <= primary_index.validate(end))

        surrogate_clauses = self._surrogate_clauses(resources)
        if len(surrogate_clauses) > 1:
            clauses.append(or_(*surrogate_clauses).self_group())
        elif len(surrogate_clauses) > 0:
            clauses.append(surrogate_clauses[0])
        return clauses

    # noinspection PyTypeChecker
    def _surrogate_clauses(self, resources: Resources) -> List[ClauseElement]:
        clauses = []

        for groups, _ in self._groupby(resources):
            group_clauses = []
            for key, value in groups.items():
                column = self.primary_key.columns[key]
                group_clauses.append(column == value)
            clauses.append(and_(*group_clauses).self_group())
        return clauses

    def extract(self, resources: Resources, result: Result[Any]) -> pd.DataFrame:
        result_columns = [r.id for r in resources]
        results = []

        # noinspection PyUnresolvedReferences
        if result.rowcount < 1:
            return pd.DataFrame(columns=result_columns)

        data = pd.DataFrame(result.fetchall(), columns=list(result.keys()))
        for group, group_resources in self._groupby(resources):

            def _is_group(row: pd.Series) -> bool:
                return all(row[col] == val for col, val in group.items())

            group_columns = self.__get_columns(group_resources)
            group_data = data[data.apply(_is_group, axis="columns")]

            for column in [c for c in group_columns if isinstance(c, DatetimeColumn)]:
                # Drop existing column, as the dtype will change from datetime64[ns] to datetime64[ns, UTC],
                # which is forbidden
                column_data = group_data[column.name].dt.tz_localize(column.timezone)
                group_data = group_data.drop(columns=[column.name])
                group_data[column.name] = column_data

            if self.datetime_index_type == DatetimeIndexType.DATE_AND_TIME:
                date_column, time_column, *_ = self.primary_key.columns
                index_column = date_column.name + time_column.name
                index_data = pd.to_datetime(group_data[date_column.name]) + group_data[time_column.name]
                group_data.loc[:, [index_column]] = index_data
                group_data.drop([date_column.name, time_column.name], inplace=True, axis="columns")
                group_data = group_data.set_index(index_column)  # .tz_localize(tz.UTC)

            elif self.datetime_index_type in (DatetimeIndexType.TIMESTAMP, DatetimeIndexType.DATETIME):
                index_column, *_ = self.primary_key.columns
                group_data = group_data.set_index(index_column.name)

            elif self.datetime_index_type == DatetimeIndexType.TIMESTAMP_UNIX:
                index_column, *_ = self.primary_key.columns
                index_data = pd.to_datetime(group_data[index_column.name], unit="s")
                group_data.drop(index_column.name, inplace=True, axis="columns")
                group_data.loc[:, [index_column.name]] = index_data
                group_data = group_data.set_index(index_column.name).tz_localize(tz.UTC)

            group_data.dropna(axis="index", how="all", inplace=True)
            group_data.rename(
                columns={r.get("column", default=r.key): r.id for r in group_resources},
                inplace=True,
            )
            if not group_data.empty:
                group_data = group_data[[r.id for r in group_resources if r.id in group_data.columns]]
                results.append(group_data)

        if len(results) == 0:
            return pd.DataFrame()
        results = sorted(results, key=lambda d: min(d.index))
        results = pd.concat(results, axis="columns")
        for result_column in [c for c in result_columns if c not in results.columns]:
            results.loc[:, [result_column]] = np.nan
        return results

    def exists(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> Select:
        columns = self.__get_columns(resources)
        select = sql.select(*columns)
        select = select.where(and_(*self._primary_clauses(resources, start, end)))
        select = select.subquery(name="exists_range")
        query = sql.select(func.count().label("count")).select_from(select)
        return query

    # noinspection PyShadowingBuiltins, PyProtectedMember, PyArgumentList
    def hash(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
    ) -> Select:
        if method.lower() not in ["md5", "sha1", "sha256", "sha512"]:
            raise ValueError(f"Invalid checksum method '{method}'")
        method = getattr(func, method.lower())
        columns = self.__get_columns(resources)

        def _validate(column: Column) -> ClauseElement:
            if column.type == DATETIME or isinstance(column.type, DATETIME):
                # TODO: Verify if there is a more generic way to implement time
                raise ValueError(
                    f"Unable to generate consistent hashes for table '{self.name}' "
                    f"with DATETIME column: {column.name}",
                )
            if column.type == TIMESTAMP or isinstance(column.type, TIMESTAMP):
                return func.unix_timestamp(column)
            if column.type == BLOB or isinstance(column.type, BLOB):
                return method(column)
            else:
                return column

        select = sql.select(*[_validate(c) for c in columns])
        nullable = [c.name for c in columns if c.nullable]
        if len(nullable) > 0:
            # Make sure to only query valid values
            select = select.filter(not_(and_(*[c.is_(None) for c in columns if c.nullable])))
        select = select.where(and_(*self._primary_clauses(resources, start, end)))
        select = select.order_by(*self._primary_order("asc"))
        select = select.subquery(name="hash_range")

        concat = func.aggregate_strings(func.concat_ws(",", *select.exported_columns.values()), ",")

        query = sql.select(method(concat).label("hash"), literal(True).label("in_range")).group_by(text("in_range"))
        return query.select_from(select)

    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        order_by: Literal["asc", "desc"] = "asc",
    ) -> Select:
        columns = self.__get_columns(resources)
        query = sql.select(*columns)
        query = query.where(and_(*self._primary_clauses(resources, start, end)))
        return query.order_by(*self._primary_order(order_by))

    # noinspection PyUnresolvedReferences
    def write(self, resources: Resources, data: pd.DataFrame) -> Insert:
        resources = resources.filter(lambda r: r.id in data.columns)
        resource_columns = self.__get_resource_columns(resources)
        primary_columns = self.primary_key.columns
        params = self._validate(resources, data.replace(np.nan, None))

        # Handle duplicate primary keys (upsert)
        if self.dialect.name == "postgresql":
            from sqlalchemy.dialects import postgresql

            query = postgresql.insert(self).values(params)
            return query.on_conflict_do_update(
                index_elements=[c.name for c in primary_columns],
                set_={c.name: c for c in resource_columns},
            )
        elif self.dialect.name in ["mariadb", "mysql"]:
            from sqlalchemy.dialects import mysql

            query = mysql.insert(self).values(params)
            return query.on_duplicate_key_update({c.name: c for c in resource_columns})
        else:
            return sql.insert(self).values(params)

    # noinspection PyMethodOverriding
    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> Delete:
        columns = self.__get_columns(resources)
        if self.columns != columns:
            raise ResourceError(f"Unable to delete rows of table '{self.name}' with only subset of columns")
        query = super().delete()
        query = query.where(and_(*self._primary_clauses(resources, start, end)))
        return query

    # noinspection PyTypeChecker
    def _validate(self, resources: Resources, data: pd.DataFrame) -> List[Dict[str, Any]]:
        values = []

        for group, group_resources in self._groupby(resources):
            group_data = data[group_resources.ids].dropna(axis="index", how="all")

            array_resources = group_resources.filter(lambda r: issubclass(r.type, pd.api.extensions.ExtensionArray))
            if len(array_resources) > 0:
                group_data = group_data.explode([r.id for r in array_resources])
            group_data.rename(columns={r.id: r.get("column", default=r.key) for r in group_resources}, inplace=True)

            for column in self.columns:
                if self.__is_datetime_index(column):
                    column_data = group_data.index
                elif column.name in group_data.columns:
                    column_data = group_data[column.name]
                elif column.name in group:
                    column_data = group[column.name]
                else:
                    column_data = None
                group_data[column.name] = column.validate(column_data)

            values.extend(group_data.to_dict(orient="records"))
        return values

    # noinspection SpellCheckingInspection
    def _groupby(self, resources: Resources) -> Iterator[Tuple[Dict[str, Any], Resources]]:
        groups: List[Tuple[Dict[str, Any], Resources]] = []

        surrogate_keys = [c for c in self.primary_key.columns if isinstance(c, SurrogateKeyColumn)]

        def _group(resource: Resource) -> Resource:
            attributes = {}
            for surrogate_key in surrogate_keys:
                if not hasattr(resource, surrogate_key.attribute):
                    raise ResourceError(
                        f"SQL resource '{resource.id}' missing surrogate key attribute: {surrogate_key.attribute}"
                    )
                attributes[surrogate_key.name] = getattr(resource, surrogate_key.attribute)
            for group in groups:
                if group[0] == attributes:
                    group[1].append(resource)
                    return resource
            groups.append((attributes, Resources([resource])))
            return resource

        resources.apply(_group)
        return iter(groups)

    def __get_columns(self, resources: Resources) -> List[Column]:
        return [*self.primary_key.columns, *self.__get_resource_columns(resources)]

    def __get_resource_columns(self, resources: Resources) -> List[Column]:
        columns = []
        for resource in resources:
            column_name = resource.get("column", default=resource.key)
            if column_name in self.columns:
                column = self.columns[column_name]
                if column not in columns + self.primary_key.columns.values():
                    columns.append(column)
        return columns

    # noinspection PyShadowingBuiltins, PyProtectedMember
    def __get_datetime_index(self, type: DatetimeIndexType) -> List[Column]:
        # Datetime index column should be the first column(s), otherwise they will be ignored
        columns = []
        for i in range(len(type.value)):
            if i >= len(self.primary_key.columns):
                break
            column = self.primary_key.columns[i]
            if column.type in type:
                columns.append(column)
        return columns

    # noinspection PyShadowingBuiltins
    def __has_datetime_index(self, *types: DatetimeIndexType) -> bool:
        return any(len(self.__get_datetime_index(t)) == len(t.value) for t in types)

    # noinspection PyShadowingBuiltins
    def __is_datetime_index(self, column: Column) -> bool:
        if self.datetime_index_type == DatetimeIndexType.NONE:
            return False
        return column in self.__get_datetime_index(self.datetime_index_type)



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .page import ComponentPage  # noqa: F401
from .group import ComponentGroup  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.columns
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .column import (  # noqa: F401
    Column,
    ColumnType,
    parse_type,
)

from .datetime import (  # noqa: F401
    DatetimeColumn,
    DatetimeType,
)

from .unix import UnixTimestampColumn  # noqa: F401

from .surrogate import SurrogateKeyColumn  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.data.channels.channel
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from copy import deepcopy
from enum import Enum
from typing import Any, Collection, Dict, Mapping, Optional, TypeVar, overload

import pandas as pd
from lories._core._configurations import Configurations
from lories._core._context import Context
from lories._core._entity import Entity, _Entity
from lories._core._resource import _Resource
from lories._core.typing import Timestamp

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class ChannelState(Enum):
    DISABLED = "DISABLED"

    DISCONNECTING = "DISCONNECTING"
    DISCONNECTED = "DISCONNECTED"
    CONNECTING = "CONNECTING"
    CONNECTED = "CONNECTED"

    VALID = "VALID"

    NOT_AVAILABLE = "NOT_AVAILABLE"

    TIMEOUT = "TIMEOUT"
    READ_ERROR = "READ_ERROR"
    WRITE_ERROR = "WRITE_ERROR"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"
    ARGUMENT_SYNTAX_ERROR = "ARGUMENT_SYNTAX_ERROR"

    def __str__(self):
        return str(self.value)


class _Channel(_Resource):
    INCLUDES: Collection[str] = (
        "logger",
        "connector",
        "converter",
        "replicator",
        "replication",
        "retention",
        "rotate",
    )
    TYPE: str = "channel"
    TIMESTAMP: str = "timestamp"

    @property
    @abstractmethod
    def freq(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def timedelta(self) -> Optional[pd.Timedelta]: ...

    @property
    @abstractmethod
    def timestamp(self) -> pd.Timestamp: ...

    @property
    @abstractmethod
    def value(self) -> Optional[Any]: ...

    @value.setter
    @abstractmethod
    def value(self, value) -> None: ...

    @property
    @abstractmethod
    def state(self) -> ChannelState | str: ...

    @state.setter
    @abstractmethod
    def state(self, state) -> None: ...

    @abstractmethod
    def is_valid(self) -> bool: ...

    @overload
    def set(
        self,
        timestamp: pd.Timestamp,
        value: Any,
    ) -> None: ...

    @overload
    def set(
        self,
        timestamp: pd.Timestamp,
        value: Any,
        state: Optional[str | ChannelState] = ChannelState.VALID,
    ) -> None: ...

    @abstractmethod
    def set(
        self,
        timestamp: pd.Timestamp,
        value: Any,
        state: Optional[str | ChannelState] = ChannelState.VALID,
    ) -> None: ...

    @abstractmethod
    def to_series(self, state: bool = False) -> pd.Series: ...

    @abstractmethod
    def from_logger(self) -> Channel: ...

    @abstractmethod
    def has_logger(self, *ids: Optional[str]) -> bool: ...

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def has_connector(self, id: Optional[str] = None) -> bool: ...

    @abstractmethod
    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None: ...

    @abstractmethod
    def read(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame: ...

    @overload
    def write(self, data: Any) -> None: ...

    @overload
    def write(self, data: pd.Series) -> None: ...

    @overload
    def write(self, data: pd.DataFrame) -> None: ...

    @abstractmethod
    def write(self, data: pd.DataFrame | pd.Series | Any) -> None: ...

    # noinspection PyShadowingBuiltins
    @classmethod
    def _build_id(
        cls,
        id: Optional[str] = None,
        key: Optional[str] = None,
        context: Optional[Context | Entity] = None,
    ) -> str:
        if id is None:
            if key is None:
                raise ValueError(f"Unable to build '{cls.__name__}' for missing ID")
            if id is None and context is not None and isinstance(context, _Entity):
                id = f"{context.id}.{key}"
            else:
                id = key
        return id

    @staticmethod
    def _build_defaults(configs: Configurations) -> Dict[str, Any]:
        return _Channel._build_configs(
            {k: deepcopy(v) for k, v in configs.items() if not isinstance(v, Mapping) or k in _Channel.INCLUDES}
        )

    @staticmethod
    def _build_configs(configs: Dict[str, Any]) -> Dict[str, Any]:
        def _build_wrapper(_key: str, _type: Optional[str] = None) -> None:
            if _type is None:
                _type = _key
            if _type not in configs:
                return
            configs[_type] = _Channel.__build_member(configs[_type], _key)

        _build_wrapper("converter")
        _build_wrapper("connector")
        _build_wrapper("connector", "logger")
        return configs

    @staticmethod
    def __build_member(configs: Optional[Dict[str, Any] | str], key: str) -> Optional[Dict[str, Any]]:
        if isinstance(configs, str) or configs is None:
            return {key: configs}
        elif not isinstance(configs, Mapping):
            raise ValueError(f"Invalid channel {key} type: " + str(configs))
        return dict(configs)


Channel = TypeVar("Channel", bound=_Channel)



# -*- coding: utf-8 -*-
"""
lories._core._configurator
~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, TypeVar

from lories._core._configurations import Configurations


class _Configurator(ABC):
    @abstractmethod
    def is_enabled(self) -> bool: ...

    @abstractmethod
    def is_configured(self) -> bool: ...

    @property
    @abstractmethod
    def configs(self) -> Optional[Configurations]: ...

    def configure(self, configs: Configurations) -> None:
        pass

    def update(self, configs: Configurations) -> None:
        pass

    @abstractmethod
    def copy(self) -> Configurator: ...

    @abstractmethod
    def duplicate(self, configs: Optional[Configurations] = None, **changes) -> Configurator: ...


Configurator = TypeVar("Configurator", bound=_Configurator)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
lories
~~~~~~

Legacy compatibility setup script for the lories package.

"""

import versioneer
from setuptools import setup

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



# -*- coding: utf-8 -*-
"""
lories.data.channels.converter
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections import OrderedDict
from typing import Any, Dict, List, Optional

import pandas as pd
from lories._core._converter import Converter, _Converter  # noqa
from lories.core.configs import ConfigurationError
from lories.core.errors import ResourceError
from lories.util import to_bool, update_recursive


class ChannelConverter:
    __configs: OrderedDict[str, Any]
    _connector: _Converter

    # noinspection PyShadowingBuiltins
    def __init__(self, converter, **configs: Any) -> None:
        if "converter" in configs:
            raise ConfigurationError("Invalid channel converter configuration 'converter'")
        self.__configs = OrderedDict(configs)
        self._converter = self._assert_converter(converter)

        self.enabled = self.__configs.pop("enabled", converter is not None)

    # noinspection PyMethodMayBeStatic
    def _assert_converter(self, converter) -> Converter:
        if converter is None or not isinstance(converter, _Converter):
            raise ResourceError(f"Invalid converter: {None if converter is None else type(converter)}")
        return converter

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ChannelConverter)
            and self._converter == other._converter
            and self._get_vars() == other._get_vars()
        )

    def __hash__(self) -> int:
        return hash((self._converter, *self._get_vars()))

    def __contains__(self, attr: str) -> bool:
        return attr in self._get_attrs()

    def __getattr__(self, attr):
        # __getattr__ gets called when the item is not found via __getattribute__
        # To avoid recursion, call __getattribute__ directly to get components dict
        configs = ChannelConverter.__getattribute__(self, f"_{ChannelConverter.__name__}__configs")
        if attr in configs.keys():
            return configs[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no configuration '{attr}'")

    def __getitem__(self, attr: str) -> Any:
        value = self.get(attr)
        if value is not None:
            return value
        raise KeyError(attr)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.key})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\tid={self.id}\n\t" + "\n\t".join(
            f"{k}={v}" for k, v in self._get_vars().items()
        )

    def __call__(self, data: Any) -> Any:
        converter_args = self._get_configs()
        if isinstance(data, pd.Series):
            converted_data = data.apply(self._converter.convert, **converter_args)
            return converted_data.apply(self._converter.to_dtype, **converter_args)
        converted_data = self._converter.convert(data, **converter_args)
        return self._converter.to_dtype(converted_data, **converter_args)

    # noinspection PyTypeChecker
    @property
    def id(self) -> str:
        return self._converter.id

    # noinspection PyTypeChecker
    @property
    def key(self) -> str:
        return self._converter.key

    def to_str(self, value: Any) -> str:
        return self._converter.to_str(value)

    def to_json(self, value: Any) -> str:
        return self._converter.to_json(value)

    def to_configs(self) -> Dict[str, Any]:
        configs = {
            "converter": self._converter.key,
            "enabled": self.enabled,
            **self._copy_configs(),
        }
        return configs

    def to_series(self, value: Any, timestamp: Optional[pd.Timestamp] = None, name: Optional[str] = None) -> pd.Series:
        return self._converter.to_series(value, timestamp=timestamp, name=name)

    def get(self, attr: str, default: Optional[Any] = None) -> Any:
        return self._get_vars().get(attr, default)

    # noinspection PyShadowingBuiltins
    def _get_vars(self) -> Dict[str, Any]:
        vars = self._copy_configs()
        vars["enabled"] = self.enabled
        return vars

    def _get_attrs(self) -> List[str]:
        return [*self._copy_configs().keys(), "enabled"]

    def _get_configs(self) -> Dict[str, Any]:
        return self.__configs

    def _copy_configs(self) -> Dict[str, Any]:
        return OrderedDict(**self._get_configs())

    def __update_configs(self, configs: Dict[str, Any]) -> None:
        update_recursive(self.__configs, configs)

    # noinspection PyShadowingBuiltins
    def _update(
        self,
        converter: Optional[str | Converter] = None,
        enabled: Optional[str | bool] = None,
        **configs: Any,
    ) -> None:
        if converter is not None:
            if isinstance(converter, str):
                if self._converter.key != converter.split(".")[-1]:
                    raise ConfigurationError(f"Unable to update channel converter from key '{converter}'")
            elif not isinstance(converter, _Converter):
                raise ConfigurationError(f"Unable to update channel converter to invalid type '{type(converter)}'")
            else:
                self._converter = converter
        if enabled is not None:
            self.enabled = to_bool(enabled)
        self.__update_configs(configs)

    def copy(self) -> ChannelConverter:
        configs = self._copy_configs()
        configs["enabled"] = self.enabled
        return type(self)(self._converter, **configs)



# -*- coding: utf-8 -*-
"""
lories.components.component
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Dict, Optional

import pandas as pd
from lories._core._component import _Component  # noqa
from lories._core._configurations import Configurations  # noqa
from lories._core._registrator import RegistratorContext  # noqa
from lories._core.typing import Timestamp  # noqa
from lories.components.access import ComponentAccess
from lories.connectors.access import ConnectorAccess
from lories.core.activator import Activator
from lories.core.register import Registrator
from lories.data.access import DataAccess
from lories.data.converters import ConverterAccess
from lories.util import to_date


class Component(_Component, Registrator, Activator):
    __converters: ConverterAccess
    __connectors: ConnectorAccess
    __components: ComponentAccess
    __data: DataAccess

    def __init__(
        self,
        context: RegistratorContext,
        configs: Optional[Configurations] = None,
        **kwargs,
    ) -> None:
        super().__init__(context=context, configs=configs, **kwargs)
        self.__converters = ConverterAccess(self)
        self.__connectors = ConnectorAccess(self)
        self.__components = ComponentAccess(self)
        self.__data = DataAccess(self)

    def _at_configure(self, configs: Configurations) -> None:
        self.__data.configure(configs.get_member(DataAccess.TYPE, ensure_exists=True))

    def _on_configure(self, configs: Configurations) -> None:
        self.__converters.load(configure=False, sort=False)
        self.__converters.sort()
        self.__converters.configure()

        self.__connectors.load(configure=False, sort=False)
        self.__connectors.sort()
        self.__connectors.configure()

        self.__components.load(configure=False, sort=False)
        self.__components.sort()
        self.__components.configure()

        self.__data.load()

    def _at_duplicate(self, **changes) -> None:
        self.__converters.duplicate(**changes)
        self.__connectors.duplicate(**changes)
        self.__components.duplicate(**changes)

    @property
    def components(self) -> ComponentAccess:
        return self.__components

    @property
    def converters(self) -> ConverterAccess:
        return self.__converters

    @property
    def connectors(self) -> ConnectorAccess:
        return self.__connectors

    @property
    def data(self) -> DataAccess:
        return self.__data

    def get(
        self,
        start: Optional[Timestamp | str] = None,
        end: Optional[Timestamp | str] = None,
        **kwargs,
    ) -> pd.DataFrame:
        start = to_date(start)
        end = to_date(end)

        data = self.__data.to_frame(unique=False)
        if data.empty or (start is not None and start < data.index[0]) or (end is not None and end > data.index[-1]):
            logged = self.__data.from_logger(start=start, end=end, unique=False)
            if not logged.empty:
                data = logged if data.empty else data.combine_first(logged)
        return self._get_range(data, start, end, **kwargs)

    @staticmethod
    def _get_range(
        data: pd.DataFrame,
        start: Optional[Timestamp | str] = None,
        end: Optional[Timestamp | str] = None,
        **kwargs,
    ) -> pd.DataFrame:
        if data.empty:
            return data
        if start is not None:
            start = to_date(start, **kwargs)
            data = data[data.index >= start]
        if end is not None:
            end = to_date(end, **kwargs)
            data = data[data.index <= end]
        return data

    # noinspection PyShadowingBuiltins
    def _get_vars(self) -> Dict[str, Any]:
        vars = super()._get_vars()
        vars.pop("type", None)
        return vars



# -*- coding: utf-8 -*-
"""
lories.data.listener
~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from collections.abc import Callable
from logging import Logger
from threading import Lock
from typing import Optional, TypeVar

import pandas as pd
import pytz as tz
from lories._core._channel import Channel  # noqa
from lories._core._channels import Channels  # noqa
from lories._core._listener import _Listener  # noqa
from lories.core import ResourceError

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


# noinspection PyShadowingBuiltins
class Listener(_Listener):
    __lock: Lock
    _logger: Logger

    _how: str
    _unique: bool

    _function: Callable[[pd.DataFrame], None]
    channels: Channels

    __start: pd.Timestamp = pd.NaT
    __complete: pd.Timestamp = pd.NaT

    def __init__(
        self,
        id: str,
        key: str,
        function: Callable[[pd.DataFrame], None],
        channels: Channels,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        super().__init__(id=id, key=key)
        self.__lock = Lock()
        self._logger = logging.getLogger(self.__module__)

        self._how = how
        self._unique = unique
        self._function = function
        self.channels = channels

    def __call__(self, timestamp: pd.Timestamp) -> Listener:
        self.__start = pd.Timestamp.now(tz=tz.UTC)
        try:
            self.__lock.acquire()
            self.run()

        except Exception as e:
            raise ListenerError(self, str(e))
        finally:
            self.__complete = timestamp
            self.__lock.release()
            self._logger.debug(f"Listener '{self.id}' finished after {round(self.runtime, 3)} seconds")
            return self

    @property
    def timestamp(self) -> pd.Timestamp | pd.NaT:
        return self.__complete

    @property
    def runtime(self) -> Optional[float]:
        if pd.isna(self.__start):
            return None
        return (pd.Timestamp.now(tz=tz.UTC) - self.__start).total_seconds()

    def run(self) -> None:
        data = self.channels.to_frame(unique=self._unique)
        self._function(data)

    def locked(self) -> bool:
        return self.__lock.locked()

    def has_update(self) -> bool:
        def _has_update(channel: Channel) -> bool:
            return channel.is_valid() and (pd.isna(self.timestamp) or self.timestamp < channel.timestamp)

        if self._how == "any":
            return any(_has_update(c) for c in self.channels)
        elif self._how == "all":
            return all(_has_update(c) for c in self.channels)
        return False


class ListenerError(ResourceError):
    """
    Raise if an error occurred notifying the listener.

    """

    # noinspection PyArgumentList
    def __init__(self, listener: ListenerType, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.listener = listener


ListenerType = TypeVar("ListenerType", bound=Listener)



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks.log
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from lories.connectors.tasks.write import WriteTask


class LogTask(WriteTask):
    def run(self) -> None:
        self._logger.debug(
            f"Logging {len(self.channels)} channels of '{type(self.connector).__name__}': {self.connector.id}"
        )
        # Pass copied connectors instead of actual objects, including parsed logger specific connector configurations
        channels = self.channels.from_logger()

        self._run_write(channels)



# -*- coding: utf-8 -*-
"""
lories.connectors.cameras.opencv
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import os

import cv2

from lories.connectors import ConnectionError, ConnectorError, register_connector_type
from lories.connectors.cameras import CameraConnector
from lories.typing import Configurations, Resources


@register_connector_type("opencv")
class OpenCV(CameraConnector):
    _capture: cv2.VideoCapture
    _host: str
    _port: int

    _username: str
    _password: str

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        self._host = configs.get("host")
        self._port = configs.get_int("port", default=554)

        self._username = configs.get("username")
        self._password = configs.get("password")

        if not all([self._host, self._port, self._username, self._password]):
            raise ValueError("Camera configuration requires 'host', 'port', 'username' and 'password'")

        os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = (
            "rtsp_transport;tcp|"  # use TCP only
            "max_delay;500000"  # 0.5 sec max internal delay
        )

        # TODO: Make timeouts configurable
        self._capture = cv2.VideoCapture()
        self._capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)
        self._capture.set(cv2.CAP_PROP_OPEN_TIMEOUT_MSEC, 3000)
        self._capture.set(cv2.CAP_PROP_READ_TIMEOUT_MSEC, 3000)

    # def is_connected(self) -> bool:
    #     return self._capture.isOpened()

    def connect(self, resources: Resources) -> None:
        super().connect(resources)
        # Validate connection only to throw ConnectionError when connect is called by the manager
        self._connect()
        self._disconnect()

    def _connect(self) -> None:
        auth = f"{self._username}:{self._password}"
        address = f"{self._host}:{self._port}/Preview_01_main"

        self._capture.open(f"rtsp://{auth}@{address}", apiPreference=cv2.CAP_FFMPEG)
        if not self._capture.isOpened():
            raise ConnectionError(self, f"Cannot open RTSP stream: 'rtsp://#:#@{address}'")

        status = False
        for _ in range(3):  # flush stale frames
            status = self._capture.grab()
        if not status:
            raise ConnectionError(self, "Failed to grab frame")

        self._logger.debug(f"Opened VideoCapture to RTSP URL 'rtsp://#:#@{address}'")

    def disconnect(self) -> None:
        super().disconnect()
        self._disconnect()

    def _disconnect(self) -> None:
        self._capture.release()
        self._logger.debug("Released VideoCapture")

    def read_frame(self) -> bytes:
        try:
            self._connect()

            status = self._capture.read()
            if not status:
                raise ConnectionError(self, "Failed to grab frame")

            status, frame = self._capture.retrieve()
            if not status or frame is None:
                raise ConnectionError(self, "Failed to retrieve frame")

            status, buffer = cv2.imencode(".jpg", frame, [int(cv2.IMWRITE_JPEG_QUALITY), 90])
            if not status:
                raise ConnectionError(self, "Failed to encode JPEG")

            return buffer.tobytes()

        except cv2.error as e:
            raise ConnectorError(self, f"OpenCV error: {e}")
        finally:
            self._disconnect()



# -*- coding: utf-8 -*-
"""
lories.components.weather.forecast
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
from typing import Optional

import pandas as pd
import pytz as tz
from lories.components.weather import Weather
from lories.core.errors import ResourceError
from lories.core.typing import Component, Configurations
from lories.location import Location
from lories.util import floor_date, to_date, to_timezone


class WeatherForecast(Weather):
    TYPE: str = "weather_forecast"
    TYPE: str = "forecast"

    interval: int = 60
    offset: int = 0

    @classmethod
    def _assert_context(cls, context: Component):
        from lories.components.weather import WeatherProvider

        if context is None or not isinstance(context, WeatherProvider):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return super()._assert_context(context)

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        self.interval = configs.get_int("interval", default=WeatherForecast.interval)
        self.offset = configs.get_int("offset", default=WeatherForecast.offset)

    def localize(self, configs: Configurations) -> None:
        # Do nothing, as context was already validated as WeatherProvider, that does have a location
        pass

    # noinspection PyUnresolvedReferences
    @property
    def location(self) -> Location:
        return self.context.location

    def get(
        self,
        start: Optional[pd.Timestamp, dt.datetime, str] = None,
        end: Optional[pd.Timestamp, dt.datetime, str] = None,
        timezone: Optional[tz.BaseTzInfo | str | int | float] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Retrieves the forecasted data for a specified time interval

        :param start:
            the start timestamp for which forecasted data will be looked up for.
            For many applications, passing datetime.datetime.now() will suffice.
        :type start:
            :class:`pandas.Timestamp`, datetime or str

        :param end:
            the end timestamp for which forecasted data will be looked up for.
        :type end:
            :class:`pandas.Timestamp`, datetime or str

        :param timezone:
            the timezone for the timestamps data will be looked up for.
        :type timezone:
            :class:`pytz.BaseTzInfo`, str or number

        :returns:
            the forecasted data, indexed in a specific time interval.

        :rtype:
            :class:`pandas.DataFrame`
        """
        forecast = self.data.to_frame(unique=False)

        # Calculate the available forecast start and end times
        if timezone is None:
            timezone = self.location.timezone
        else:
            timezone = to_timezone(timezone)
        end = to_date(end, timezone=timezone)
        start = to_date(start, timezone=timezone)
        if start is None:
            start = pd.Timestamp.now(tz=timezone)

        if forecast.empty or start < forecast.index[0] or end > forecast.index[-1]:
            start_schedule = floor_date(start, self.location.timezone, freq=f"{self.interval}T")
            start_schedule += pd.Timedelta(minutes=self.offset)
            if start_schedule > start:
                start_schedule -= pd.Timedelta(minutes=self.interval)

            logged = self.data.from_logger(start=start, end=end, unique=False)
            if not logged.empty:
                forecast = logged if forecast.empty else forecast.combine_first(logged)

        return self._get_range(forecast, start, end, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.simulation.durations
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
import json
import os
from collections.abc import Mapping
from typing import Iterator

from lories import Directory


class Durations(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self, dir: Directory) -> None:
        self._durations = {}
        self._file = os.path.join(dir, "durations.json")
        if os.path.isfile(self._file):
            with open(self._file, "r", encoding="utf-8") as f:
                self._durations = json.load(f)
                for duration in self._durations.values():

                    def get_duration(key):
                        return dt.datetime.strptime(duration[key], "%Y-%m-%d %H:%M:%S.%f")

                    if "start" in duration:
                        duration["start"] = get_duration("start")
                    if "end" in duration:
                        duration["end"] = get_duration("end")

    def __repr__(self) -> str:
        return str(self._durations)

    def __iter__(self) -> Iterator:
        return iter(self._durations)

    def __len__(self) -> int:
        return len(self._durations)

    def __getitem__(self, key: str) -> float:
        return self._durations[key]["minutes"]

    def start(self, key: str) -> None:
        if key not in self._durations:
            self._durations[key] = {}
        if "minutes" not in self._durations[key]:
            self._durations[key]["minutes"] = 0
        if "complete" not in self._durations[key]:
            self._durations[key]["complete"] = False
        if not self._durations[key]["complete"]:
            if "end" in self._durations[key]:
                del self._durations[key]["end"]
            self._durations[key]["start"] = dt.datetime.now()

    def stop(self, key: str = None) -> None:
        for k in self.keys() if key is None else [key]:
            self._stop(k)
        self._write()

    def _stop(self, key: str = None) -> None:
        if key not in self._durations:
            raise ValueError('No duration found for key: "{}"'.format(key))
        if "start" not in self._durations[key]:
            raise ValueError('Timer for key "{}" not started yet'.format(key))

        self._durations[key]["end"] = dt.datetime.now()

        minutes = self._durations[key]["minutes"] if "minutes" in self._durations[key] else 0
        minutes += round((self._durations[key]["end"] - self._durations[key]["start"]).total_seconds() / 60.0, 6)
        self._durations[key]["minutes"] = minutes

    # noinspection PyTypeChecker
    def _write(self) -> None:
        with open(self._file, "w", encoding="utf-8") as file:
            json.encoder.FLOAT_REPR = lambda o: format(o, ".3f")
            json.dump(self._durations, file, indent=4, default=str, ensure_ascii=False)

    def complete(self) -> None:
        for key in self.keys():
            self._durations[key]["complete"] = True
            self.stop(key)

    def is_complete(self) -> bool:
        return all(d["complete"] for d in self._durations.values())



# -*- coding: utf-8 -*-
"""
lories.core.resources
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from typing import Any, Callable, Dict, Generic, Iterable, Iterator, List, Sequence, Tuple

from lories._core._resource import Resource  # noqa
from lories._core._resources import Resources as ResourcesType  # noqa
from lories._core._resources import _Resources  # noqa


class Resources(_Resources, Generic[Resource]):
    _resources: List[Resource]

    def __init__(self, resources=()) -> None:
        self._logger = logging.getLogger(type(self).__module__)
        self._resources = [*resources]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(str(r.id) for r in self._resources)})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{r.id} = {repr(r)}" for r in self._resources)

    def __contains__(self, resource: str | Resource) -> bool:
        if isinstance(resource, str):
            return any(resource == r.id for r in self._resources)
        return resource in self._resources

    def __getitem__(self, index: Iterable[str] | str | int) -> Resource | ResourcesType:
        if isinstance(index, str):
            for resource in self._resources:
                if resource.id == index:
                    return resource
        if isinstance(index, Iterable):
            return type(self)([r for r in self._resources if r.id == index])
        raise KeyError(index)

    def __iter__(self) -> Iterator[Resource]:
        return iter(self._resources)

    def __len__(self) -> int:
        return len(self._resources)

    def __add__(self, other) -> ResourcesType:
        return type(self)([*self, *other])

    def append(self, resource: Resource) -> None:
        self._resources.append(resource)

    def extend(self, resources: Iterable[Resource]) -> None:
        self._resources.extend(resources)

    def update(self, resources: Iterable[Resource]) -> None:
        resource_ids = [r.id for r in resources]
        for resource in [r for r in self._resources if r.id in resource_ids]:
            self._resources.remove(resource)
        self._resources.extend(resources)

    @property
    def ids(self) -> Sequence[str]:
        return [str(resource.id) for resource in self._resources]

    @property
    def keys(self) -> Sequence[str]:
        return [str(resource.key) for resource in self._resources]

    def copy(self) -> ResourcesType:
        return type(self)([resource.copy() for resource in self._resources])

    def apply(self, apply: Callable[[Resource], Resource], inplace: bool = False) -> ResourcesType:
        resources = self._resources if not inplace else self._resources.copy()
        return type(self)([apply(resource) for resource in resources])

    def filter(self, *filters: Callable[[Resource], bool]) -> ResourcesType:
        def _filters(channel: Resource) -> bool:
            for _filter in filters:
                if _filter is not None and not _filter(channel):
                    return False
            return True

        return type(self)([resource for resource in self._resources if _filters(resource)])

    # noinspection PyShadowingBuiltins, SpellCheckingInspection
    def groupby(self, by: Callable[[Resource], Any] | str) -> Iterator[Tuple[Any, ResourcesType]]:
        def _by(r: Resource) -> Any:
            return r.get(by, default=None)

        filter = _by if isinstance(by, str) else by
        for group_by in list(dict.fromkeys([filter(r) for r in self])):
            yield group_by, self.filter(lambda r: filter(r) == group_by)

    # noinspection PyTypeChecker
    def to_configs(self) -> Dict[str, Any]:
        return {r.id: r.to_configs() for r in self._resources}



# -*- coding: utf-8 -*-
"""
lories.data.channels.channels
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Callable
from typing import Any, Iterator, Tuple

import numpy as np
import pandas as pd
from lories._core._channel import Channel, ChannelState, _Channel  # noqa
from lories._core._channels import Channels as ChannelsType  # noqa
from lories._core._channels import _Channels  # noqa
from lories.core import Resources
from lories.data.validation import validate_index

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class Channels(_Channels, Resources[Channel]):
    def __str__(self) -> str:
        return str(self.to_frame(unique=True, states=True))

    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        for channel in self:
            channel.register(function, how=how, unique=unique)

    # noinspection PyTypeChecker
    def apply(self, apply: Callable[[Channel], Channel], inplace: bool = False) -> ChannelsType:
        return super().apply(apply, inplace=inplace)

    def filter(self, *filters: Callable[[Channel], bool]) -> ChannelsType:
        return super().filter(*filters)

    # noinspection PyShadowingBuiltins, SpellCheckingInspection
    def groupby(self, by: Callable[[Channel], Any] | str) -> Iterator[Tuple[Any, ChannelsType]]:
        return super().groupby(by)

    def from_logger(self) -> ChannelsType:
        return type(self)([c.from_logger() for c in self if c.has_logger()])

    def to_frame(self, unique: bool = False, states: bool = False) -> pd.DataFrame:
        data = OrderedDict()
        columns = list(self.keys if not unique else self.ids)
        for channel in self:
            if pd.isna(channel.timestamp):
                continue
            channel_uid = channel.key if not unique else channel.id
            channel_data = channel.to_series(state=states)
            channel_data.name = channel_uid
            if channel_data.empty:
                continue
            for timestamp, channel_values in channel_data.to_frame().to_dict(orient="index").items():
                if timestamp not in data:
                    timestamp_data = data[timestamp] = {c: np.nan for c in columns}
                else:
                    timestamp_data = data[timestamp]

                    if any(not pd.isna(timestamp_data[k]) for k in channel_values.keys()):
                        self._logger.warning(
                            f"Overriding value for duplicate index while merging channel '{channel.id}' into "
                            f"DataFrame for index: {channel_data.index}"
                        )
                timestamp_data.update(channel_values)

        if len(data) == 0:
            return pd.DataFrame(columns=columns)
        data = pd.DataFrame.from_records(
            data=list(data.values()),
            index=list(data.keys()),
        )
        data.dropna(axis="index", how="all", inplace=True)
        data = validate_index(data)
        data.index.name = _Channel.TIMESTAMP
        return data

    # noinspection PyProtectedMember
    def set_frame(self, data: pd.DataFrame) -> None:
        for converter, channels in self.groupby(lambda c: c.converter._converter):
            converted_data = converter.from_frame(data, channels)
            for channel in channels:
                channel_data = converted_data.loc[:, channel.id].dropna()
                if channel_data.empty or channel_data.isna().all():
                    channel.state = ChannelState.NOT_AVAILABLE
                    self._logger.debug(f"Missing value for channel: {channel.id}")
                    continue

                timestamp = channel_data.index[0]
                if len(channel_data.index) == 1:
                    channel_data = channel_data.values[0]
                channel._set(timestamp, channel_data, ChannelState.VALID)

    def set_state(self, state: ChannelState) -> None:
        def _set_state(channel: Channel) -> Channel:
            channel.state = state
            return channel

        self.apply(_set_state, inplace=True)



# -*- coding: utf-8 -*-
"""
lories._core._database
~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Optional, TypeVar, overload

import pandas as pd
from lories._core._connector import _Connector
from lories._core._resources import Resources
from lories._core.typing import Timestamp

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class _Database(_Connector):
    # noinspection PyShadowingBuiltins
    @abstractmethod
    def hash(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
        encoding: str = "UTF-8",
    ) -> Optional[str]: ...

    def exists(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> bool: ...

    @overload
    def read(self, resources: Resources) -> pd.DataFrame: ...

    @overload
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def read_first(self, resources: Resources) -> Optional[pd.DataFrame]: ...

    @abstractmethod
    def read_first_index(self, resources: Resources) -> Optional[Any]: ...

    @abstractmethod
    def read_last(self, resources: Resources) -> Optional[pd.DataFrame]: ...

    @abstractmethod
    def read_last_index(self, resources: Resources) -> Optional[pd.Index]: ...

    @overload
    def delete(self, resources: Resources) -> None: ...

    @overload
    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None: ...

    @abstractmethod
    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None: ...


Database = TypeVar("Database", bound=_Database)



# -*- coding: utf-8 -*-
"""
lories.core.register.registry
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import builtins
from typing import Callable, Generic, List, Optional, Type

from lories._core._configurations import Configurations  # noqa
from lories._core._registrator import Registrator, RegistratorContext  # noqa
from lories.core.errors import ResourceError


# noinspection PyShadowingBuiltins
class Registration(Generic[Registrator]):
    __class: Type[Registrator]
    __factory: RegistrationFactory

    _key: str
    alias: List[str]

    def __init__(
        self,
        cls: Type[Registrator],
        type: str,
        *alias: str,
        factory: Optional[RegistrationFactory] = None,
    ):
        if not isinstance(type, str):
            raise ResourceError(f"Invalid '{builtins.type(type)}' registration type: {type}")
        self._key = type.lower()
        self.alias = list(a.lower() for a in alias if a is not None and isinstance(a, str))
        self.__class = cls

        if factory is not None:
            if not callable(factory):
                raise ResourceError(f"Invalid registration initialization function: {factory}")
            self.__factory = factory
        else:
            self.__factory = cls

    @property
    def key(self) -> str:
        return self._key

    @property
    def name(self) -> str:
        return self.__class.__name__

    @property
    def type(self) -> Type[Registrator]:
        return self.__class

    def is_type(self, type: str) -> bool:
        return self.key == type or self.is_alias(type)

    def is_alias(self, type: str) -> bool:
        return any(type.startswith(a) for a in self.alias)

    def is_instance(self, registrator: Registrator) -> bool:
        return isinstance(registrator, self.__class)

    def initialize(self, *args, **kwargs) -> Registrator:
        # registrator = self.__factory(*args, **kwargs)
        # if not isinstance(registrator, self.__class):
        #     raise ResourceError(f"{self.name} factory instanced invalid '{type(registrator)}' registrator")
        return self.__factory(*args, **kwargs)


class RegistrationError(ResourceError):
    """
    Raise if an error with the registration occurred.

    """


RegistrationFactory = Callable[[RegistratorContext, Optional[Configurations]], Registrator]



# -*- coding: utf-8 -*-
"""
lories.data.access
~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Callable, Collection, Iterable, Optional, Type, overload

import pandas as pd
from lories._core import _Context, _Registrator  # noqa
from lories._core._data import DataContext, _DataContext, _DataManager  # noqa
from lories.core import Configurator, Constant, ResourceError
from lories.core.typing import ChannelsArgument, Registrator, Timestamp
from lories.data.channels import Channel, Channels
from lories.data.context import DataContext as _DataAccess
from lories.util import get_context, update_recursive

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


# noinspection PyProtectedMember, PyShadowingBuiltins
class DataAccess(_DataAccess, Configurator):
    __registrar: _Registrator
    __context: _DataContext

    def __init__(self, registrar: Registrator, **kwargs: Any) -> None:
        registrar = self._assert_registrar(registrar)
        super().__init__(logger=registrar._logger, **kwargs)
        self.__registrar = registrar
        self.__context = self._assert_context(get_context(registrar, _DataManager))

    @classmethod
    def _assert_registrar(cls, registrar: Registrator) -> Registrator:
        if registrar is None or not isinstance(registrar, _Registrator):
            raise TypeError(f"Invalid '{cls.__name__}' registrator: {type(registrar)}")
        return registrar

    @classmethod
    def _assert_context(cls, context: DataContext) -> DataContext:
        if context is None or not isinstance(context, _DataManager):
            raise TypeError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(str(c.key) for c in self.values())})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{v.key} = {repr(v)}" for v in self.values())

    def __getitem__(self, id: Iterable[str] | str) -> Channel | Channels:
        if isinstance(id, str):
            return self._get(id)
        if isinstance(id, Iterable):
            return Channels([self._get(i) for i in id])
        raise KeyError(id)

    # noinspection PyArgumentList
    def __getattr__(self, attr):
        channels = _Context.__getattribute__(self, f"{_Context.__name__}__map")
        channels_by_key = {c.key: c for c in channels.values()}
        if attr in channels_by_key:
            return channels_by_key[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def __validate_id(self, id: str) -> str:
        if not len(id.split(".")) > 1:
            id = f"{self.__registrar.id}.{id}"
        return id

    # noinspection PyArgumentList
    def _contains(self, __channel: str | Channel) -> bool:
        channels = _Context.__getattribute__(self, f"{_Context.__name__}__map")
        if isinstance(__channel, str):
            __channel = self.__validate_id(__channel)
            return __channel in channels.keys()
        if isinstance(__channel, _Registrator):
            return __channel in channels.values()
        return False

    def _get(self, id: str) -> Channel:
        return super()._get(self.__validate_id(id))

    def _set(self, id: str, channel: Channel) -> None:
        id = self.__validate_id(id)

        self.context._set(id, channel)
        super()._set(id, channel)

    def _create(self, id: str, key: str, type: Type, **configs: Any) -> Channel:
        return self.context._create(id=id, key=key, type=type, **configs)

    def _remove(self, *__objects: str | Channel) -> None:
        for __object in __objects:
            if isinstance(__object, str):
                __object = self.__validate_id(__object)

            self.context._remove(__object)
            super()._remove(__object)

    @property
    def empty(self) -> bool:
        return len(self.values()) == 0 or self.to_frame(states=False).dropna(axis="index", how="all").empty

    @property
    def context(self) -> DataContext:
        return self.__context

    def load(self, sort: bool = True) -> Collection[Channel]:
        channels = []
        defaults = {}
        if self.configs.has_member(Channels.TYPE):
            configs = self.configs.get_member(Channels.TYPE)
            defaults = Channel._build_defaults(configs)
            channels.extend(self._load_from_members(self.__registrar, configs))
        channels.extend(
            self._load_from_file(self.__registrar, self.configs.dirs, f"{Channels.TYPE}.conf", defaults=defaults)
        )

        if sort:
            self.sort()
        return channels

    def add(self, key: str | Constant, **configs: Any) -> None:
        if isinstance(key, Constant):
            configs = {
                **key.to_dict(),
                **configs,
            }
            key = configs.pop("key")
        configs = Channel._build_configs(configs)
        channels = self.configs.get_member(Channels.TYPE, ensure_exists=True)
        if not channels.has_member(key):
            channels._add_member(key, configs)
        else:
            channel_configs = Channel._build_configs(channels[key])
            channel_configs = update_recursive(channel_configs, configs, replace=False)
            channels[key] = channel_configs

        if self.__registrar.is_configured():
            channel_defaults = Channel._build_defaults(channels)
            channel_configs = Channel._build_configs(channel_defaults)
            # Be wary of the order. First, update the channel core with the default core
            # of the configuration file, then update the function arguments. Last, override
            # everything with the channel specific configurations of the file.
            channel_configs = update_recursive(channel_configs, configs)
            channel_configs = update_recursive(channel_configs, channels[key])
            channel_id = f"{self.__registrar.id}.{key}"
            if self._contains(channel_id):
                self._update(id=channel_id, key=key, **channel_configs)
            else:
                channel = self._create(id=channel_id, key=key, **channel_configs)
                self._add(channel)

    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        channels: Optional[ChannelsArgument] = None,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        channels = self._filter_by_args(channels)
        self.__context.register(function, channels=channels, how=how, unique=unique)

    def has_logged(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> bool:
        channels = self._filter_by_args(channels)
        return self.__context.has_logged(channels=channels, start=start, end=end, timeout=timeout)

    @overload
    def from_logger(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
        unique: bool = False,
    ) -> pd.DataFrame: ...

    @overload
    def from_logger(
        self,
        channels: ChannelsArgument,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
        unique: bool = False,
    ) -> pd.DataFrame: ...

    def from_logger(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
        unique: bool = False,
    ) -> pd.DataFrame:
        return self.read_logged(channels=channels, start=start, end=end, timeout=timeout, unique=unique)

    def read_logged(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
        unique: bool = False,
    ) -> pd.DataFrame:
        channels = self._filter_by_args(channels)
        data = self.__context.read_logged(channels=channels, start=start, end=end, timeout=timeout)
        if not unique:
            data.rename(columns={c.id: c.key for c in channels}, inplace=True)
        return data

    def read(
        self,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
        unique: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        channels = self._filter_by_args(channels)
        data = self.__context.read(channels=channels, timeout=timeout, **kwargs)
        if not unique:
            data.rename(columns={c.id: c.key for c in channels}, inplace=True)
        return data

    def write(
        self,
        data: pd.DataFrame,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
    ) -> None:
        if data is None:
            raise ResourceError(f"Invalid data to write '{self.id}': {data}")
        data.rename(columns={c.key: c.id for c in channels}, inplace=True)
        channels = self._filter_by_args(channels)
        self.__context.write(data, channels=channels, timeout=timeout)



# -*- coding: utf-8 -*-
"""
lories.data
~~~~~~~~~~~


"""

from . import converters  # noqa: F401
from .converters import (  # noqa: F401
    Converter,
    ConversionError,
)

from . import channels  # noqa: F401
from .channels import (  # noqa: F401
    ChannelState,
    Channel,
    Channels,
)

from . import listeners  # noqa: F401
from .listeners import (  # noqa: F401
    Listener,
    ListenerError,
)

from .context import DataContext  # noqa: F401

from .access import DataAccess  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.connectors.cameras.camera
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

from abc import abstractmethod
from time import sleep, time
from typing import Iterable

import pandas as pd
from lories.connectors import ConnectionError, Connector
from lories.typing import Resources


class CameraConnector(Connector):
    def read(self, resources: Resources) -> pd.DataFrame:
        timestamp = pd.Timestamp.now(tz="UTC").floor(freq="s")

        # TODO: Wrap read_frame() and cache latest frame to only read if frame is older than a second
        data = self.read_frame()
        return pd.DataFrame(data=[data] * len(resources), index=[timestamp], columns=list(resources.ids))

    @abstractmethod
    def read_frame(self) -> bytes: ...

    def stream(self, fps: int = 30) -> Iterable[bytes]:
        while True:
            try:
                now = time()

                if self.is_connected():
                    yield self.read_frame()

                seconds = (1 / fps) - (time() - now)
                if seconds > 0:
                    sleep(seconds)

            except KeyboardInterrupt:
                pass
            except ConnectionError as e:
                self._logger.error(f"Unexpected error '{e}' while streaming")
                self.disconnect()

    def write(self, data: pd.DataFrame) -> None:
        raise NotImplementedError("Camera connector does not support writing")



# -*- coding: utf-8 -*-
"""
lories._core._component
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Collection, Dict, Optional, TypeAlias, TypeVar, overload

import pandas as pd
from lories._core._activator import _Activator
from lories._core._channel import _Channel
from lories._core._channels import _Channels
from lories._core._configurations import Configurations
from lories._core._connector import Connectors
from lories._core._converter import Converters
from lories._core._data import Data, _DataContext
from lories._core._registrator import _Registrator, _RegistratorContext
from lories._core.typing import Timestamp


class _Component(_Registrator, _Activator):
    INCLUDES: Collection[str] = (_DataContext.TYPE,)
    TYPE: str = "component"

    @property
    @abstractmethod
    def components(self) -> Components: ...

    @property
    @abstractmethod
    def converters(self) -> Converters: ...

    @property
    @abstractmethod
    def connectors(self) -> Connectors: ...

    @property
    @abstractmethod
    def data(self) -> Data: ...

    @overload
    def get(self) -> pd.DataFrame: ...

    @overload
    def get(self, date: Timestamp | str) -> pd.DataFrame: ...

    @overload
    def get(
        self,
        start: Timestamp | str,
        end: Timestamp | str,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def get(
        self,
        start: Optional[Timestamp | str] = None,
        end: Optional[Timestamp | str] = None,
        **kwargs,
    ) -> pd.DataFrame: ...

    # noinspection PyProtectedMember
    @classmethod
    def _build_defaults(
        cls,
        configs: Configurations,
        includes: Optional[Collection[str]] = (),
        strict: bool = False,
    ) -> Dict[str, Any]:
        defaults = super()._build_defaults(configs, includes)
        if strict and _DataContext.TYPE in defaults:
            defaults[_DataContext.TYPE][_Channels.TYPE] = _Channel._build_defaults(
                defaults[_DataContext.TYPE].get_member(_Channels.TYPE, defaults={})
            )
        return defaults


Component = TypeVar("Component", bound=_Component)


# noinspection PyAbstractClass
class _ComponentContext(_RegistratorContext[Component]):
    TYPE: str = "components"


ComponentContext = TypeVar(
    name="ComponentContext",
    bound=_ComponentContext,
)
Components: TypeAlias = ComponentContext



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.columns.timestamp
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
from typing import Any, Optional

from sqlalchemy.types import INTEGER

import numpy as np
import pandas as pd
import pytz as tz
from lories.connectors.sql.columns import Column, ColumnType


# noinspection PyShadowingBuiltins
class UnixTimestampColumn(Column):
    inherit_cache: bool = True

    def __init__(
        self,
        name: str,
        type: ColumnType = INTEGER,
        *args,
        default: Optional[Any] = None,
        nullable: bool = False,
        **kwargs,
    ) -> None:
        # if default is None and not nullable:
        #     default = "(UNIX_TIMESTAMP())"
        super().__init__(
            name,
            type,
            *args,
            default=default,
            nullable=nullable,
            primary_key=True,
            **kwargs,
        )

    # noinspection PyUnresolvedReferences
    def validate(self, data: Any) -> Any:
        if isinstance(data, pd.Series):
            data = data.dt.tz_convert(tz.UTC).view(np.int64) // 10**9
        elif isinstance(data, pd.DatetimeIndex):
            data = data.tz_convert(tz.UTC).view(np.int64) // 10**9
        elif isinstance(data, pd.Timestamp):
            epoch = dt.datetime(1970, 1, 1, tzinfo=tz.UTC)
            data = data.tz_convert(tz.UTC)
            data = int((data - epoch).total_seconds())
        else:
            epoch = dt.datetime(1970, 1, 1, tzinfo=tz.UTC)
            data = data.astimezone(tz.UTC)
            data = int((data - epoch).total_seconds())
        return super().validate(data)



# -*- coding: utf-8 -*-
"""
lories.data.channels.connector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections import OrderedDict
from typing import Any, Dict, List, Optional

import pandas as pd
from lories._core._connector import Connector, _Connector  # noqa
from lories._core._database import _Database  # noqa
from lories.core.configs import ConfigurationError
from lories.core.errors import ResourceError
from lories.util import to_bool, update_recursive


class ChannelConnector:
    __configs: OrderedDict[str, Any]
    _connector: Optional[Connector]

    enabled: bool = False

    timestamp: pd.Timestamp = pd.NaT

    # noinspection PyShadowingBuiltins
    def __init__(self, connector: Optional[Connector] = None, **configs: Any) -> None:
        if "connector" in configs:
            raise ConfigurationError("Invalid channel connector configuration 'connector'")
        self.__configs = OrderedDict(configs)
        self._connector = self._assert_connector(connector)

        self.enabled = to_bool(self.__configs.pop("enabled", connector is not None and connector.is_enabled()))

    @classmethod
    def _assert_connector(cls, connector: Connector) -> Optional[Connector]:
        if connector is None:
            return None
        if not isinstance(connector, _Connector):
            raise ResourceError(f"Invalid connector: {None if connector is None else type(connector)}")
        return connector

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ChannelConnector)
            and self._connector == other._connector
            and self._get_vars() == other._get_vars()
        )

    def __hash__(self) -> int:
        return hash((self._connector, *self._get_vars()))

    def __contains__(self, attr: str) -> bool:
        return attr in self._get_attrs()

    def __getattr__(self, attr):
        # __getattr__ gets called when the item is not found via __getattribute__
        # To avoid recursion, call __getattribute__ directly to get components dict
        configs = ChannelConnector.__getattribute__(self, f"_{ChannelConnector.__name__}__configs")
        if attr in configs.keys():
            return configs[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no configuration '{attr}'")

    def __getitem__(self, attr: str) -> Any:
        value = self.get(attr)
        if value is not None:
            return value
        raise KeyError(attr)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.id})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\tid={self.id}\n\t" + "\n\t".join(
            f"{k}={v}" for k, v in self._get_vars().items()
        )

    @property
    def id(self) -> Optional[str]:
        return self._connector.id if self._connector is not None else None

    @property
    def key(self) -> Optional[str]:
        return self._connector.key if self._connector is not None else None

    def is_configured(self) -> bool:
        return self._connector.is_configured() if self.enabled else False

    def is_connected(self) -> bool:
        return self._connector.is_connected() if self.enabled else False

    def is_database(self) -> bool:
        return self._is_database(self._connector) if self.enabled else False

    @staticmethod
    def _is_database(connector: Connector) -> bool:
        return isinstance(connector, _Database)

    def to_configs(self) -> Dict[str, Any]:
        configs = {
            "connector": self._connector.key if self._connector else None,
            "enabled": self.enabled,
            **self._copy_configs(),
        }
        return configs

    def get(self, attr: str, default: Optional[Any] = None) -> Any:
        return self._get_vars().get(attr, default)

    # noinspection PyShadowingBuiltins
    def _get_vars(self) -> Dict[str, Any]:
        vars = self._get_args()
        vars["timestamp"] = self.timestamp
        return vars

    # noinspection PyShadowingBuiltins
    def _get_args(self) -> Dict[str, Any]:
        args = self._copy_configs()
        args["enabled"] = self.enabled
        return args

    def _get_attrs(self) -> List[str]:
        return [*self._copy_configs().keys(), "timestamp", "enabled"]

    def _get_configs(self) -> Dict[str, Any]:
        return self.__configs

    def _copy_configs(self) -> Dict[str, Any]:
        return OrderedDict(**self._get_configs())

    def __update_configs(self, configs: Dict[str, Any]) -> None:
        update_recursive(self.__configs, configs)

    # noinspection PyShadowingBuiltins
    def _update(
        self,
        connector: Optional[str | Connector] = None,
        enabled: Optional[str | bool] = None,
        **configs: Any,
    ) -> None:
        if connector is not None:
            if isinstance(connector, str):
                if self._connector is None or self._connector.key != connector.split(".")[-1]:
                    raise ConfigurationError(f"Unable to update channel connector from key '{connector}'")
            elif not isinstance(connector, _Connector):
                raise ConfigurationError(f"Unable to update channel connector to invalid type '{type(connector)}'")
            else:
                self._connector = connector
        if enabled is not None:
            self.enabled = to_bool(enabled)
        self.__update_configs(configs)

    def copy(self) -> ChannelConnector:
        return type(self)(self._connector, **self._get_args())



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks.write
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from lories._core._channels import Channels  # noqa
from lories.connectors.tasks.task import ConnectorTask


class WriteTask(ConnectorTask):
    def run(self) -> None:
        self._logger.debug(
            f"Writing {len(self.channels)} channels of '{type(self.connector).__name__}': {self.connector.id}"
        )
        self._run_write(self.channels)

    def _run_write(self, channels: Channels) -> None:
        self.connector.write(channels.to_frame(unique=True))



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.header
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from os import PathLike
from typing import List, Optional

from dash import html
from dash.development.base_component import Component
from dash_bootstrap_components import Button, Col, Input, NavbarBrand, NavbarSimple, Row
from PIL import Image


class PageHeader:
    _brand: Component

    menu: List[Component]

    def __init__(self, name: str, logo: Optional[str | PathLike] = None) -> None:
        super().__init__()
        if logo is not None:
            self._brand = Row(
                [
                    Col(html.Img(src=Image.open(logo), height="30px")),
                    Col(NavbarBrand(name, className="ms-2")),
                ],
                align="center",
                className="g-0",
            )
        else:
            self._brand = NavbarBrand(name, className="ms-2")

        self.menu = []

    @property
    def searchbar(self) -> Row:
        return Row(
            [
                Col(Input(type="search", placeholder="Search")),
                Col(
                    Button("Search", color="primary", className="ms-2", n_clicks=0),
                    width="auto",
                ),
            ],
            className="g-0 ms-auto flex-nowrap mt-3 mt-md-0",
            align="center",
        )

    @property
    def navbar(self) -> NavbarSimple:
        return NavbarSimple(
            children=self.menu,
            brand=self._brand,
            brand_href="/",
            links_left=True,
            color="light",
            dark=False,
        )



# -*- coding: utf-8 -*-
"""
lories.core.register.access
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Dict, Generic, Optional, Sequence, overload

from lories._core._context import Context, _Context  # noqa
from lories._core._registrator import (  # noqa
    Registrator,
    RegistratorContext,
    _Registrator,  # noqa
    _RegistratorContext,  # noqa
)
from lories.core.configs import Configurations
from lories.core.errors import ResourceError
from lories.core.register._load import _RegistratorLoader
from lories.util import update_recursive


# noinspection SpellCheckingInspection, PyProtectedMember, PyAbstractClass, PyShadowingBuiltins
class RegistratorAccess(_RegistratorLoader[Registrator], Generic[Registrator]):
    _registrar: Registrator
    __context: RegistratorContext

    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        context: RegistratorContext,
        registrar: Registrator,
        **kwargs,
    ) -> None:
        context = self._assert_context(context)
        registrar = self._assert_registrar(registrar)
        super().__init__(logger=registrar._logger, **kwargs)
        self.__context = context
        self._registrar = registrar

    @classmethod
    def _assert_registrar(cls, registrar: Registrator):
        if registrar is None or not isinstance(registrar, _Registrator):
            raise ResourceError(f"Invalid '{cls.__name__}' registrator: {type(registrar)}")
        return registrar

    @classmethod
    def _assert_context(cls, context: RegistratorContext):
        if context is None or not isinstance(context, _RegistratorContext):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(str(c.key) for c in self.values())})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{v.key} = {repr(v)}" for v in self.values())

    # noinspection PyArgumentList
    def __getattr__(self, attr):
        registrators = _Context.__getattribute__(self, f"{_Context.__name__}__map")
        registrators_by_key = {c.key: c for c in registrators.values()}
        if attr in registrators_by_key:
            return registrators_by_key[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    @property
    def context(self) -> RegistratorContext:
        return self.__context

    def __validate_id(self, id: str) -> str:
        if not len(id.split(".")) > 1:
            id = f"{self._registrar.id}.{id}"
        return id

    # noinspection PyArgumentList
    def _contains(self, __registrator: str | Registrator) -> bool:
        registrators = _Context.__getattribute__(self, f"{_Context.__name__}__map")
        if isinstance(__registrator, str):
            __registrator = self.__validate_id(__registrator)
            return __registrator in registrators.keys()
        if isinstance(__registrator, _Registrator):
            return __registrator in registrators.values()
        return False

    def _get(self, id: str) -> Registrator:
        return super()._get(self.__validate_id(id))

    def _set(self, id: str, registrator: Registrator) -> None:
        id = self.__validate_id(id)

        self.__context._set(id, registrator)
        super()._set(id, registrator)

    def _create(
        self,
        context: Context | Registrator,
        configs: Configurations,
        **kwargs: Any,
    ) -> Registrator:
        return self.context._create(context, configs, **kwargs)

    def _remove(self, *__registrators: str | Registrator) -> None:
        for __registrator in __registrators:
            if isinstance(__registrator, str):
                __registrator = self.__validate_id(__registrator)

            self.__context._remove(__registrator)
            super()._remove(__registrator)

    @overload
    def add(self, key: str, **configs: Any) -> None: ...

    @overload
    def add(self, registrator: Registrator) -> None: ...

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def add(self, __registrator: str | Registrator, **configs: Any) -> None:
        if isinstance(__registrator, _Registrator):
            self._add(__registrator)
            return

        registrators = self._load_registrators_configs()
        if not registrators.has_member(__registrator):
            registrators._add_member(__registrator, configs)
        else:
            registrator_configs = registrators[__registrator]
            registrator_configs = update_recursive(registrator_configs, configs, replace=False)
            registrators[__registrator] = registrator_configs

        if self._registrar.is_configured():
            registrator_configs = Configurations.load(
                f"{__registrator}.conf",
                **registrators.dirs.to_dict(),
                **self._build_defaults(registrators),
            )
            # Be wary of the order. First, update the registrator core with the default core
            # of the configuration file, then update the function arguments. Last, override
            # everything with the registrator specific configurations of the file.
            registrator_configs = update_recursive(registrator_configs, configs)
            registrator_configs = update_recursive(registrator_configs, registrators[__registrator])
            registrator = self._load_from_configs(self._registrar, registrator_configs)
            if registrator.is_enabled():
                registrator.configure(registrator_configs)

    # noinspection PyUnresolvedReferences
    def duplicate(self, configs: Configurations, **_) -> None:
        for registrator in self.values():
            relative_dir = registrator.configs.dirs.conf.relative_to(registrator.context.configs.dirs.conf)
            registrator_dirs = configs.dirs.copy()
            registrator_dirs.conf = registrator_dirs.conf.joinpath(relative_dir)
            registrator.configs.copy(registrator_dirs)

    def _load_registrator_defaults(self, configs: Optional[Configurations] = None, **kwargs) -> Dict[str, Any]:
        _class = self._get_registator_class()
        if configs is None:
            configs = self._registrar.configs
        return _class._build_defaults(configs, **kwargs)

    # noinspection PyUnresolvedReferences
    def _load_registrators_configs(self) -> Configurations:
        _type = self._get_registrators_type()
        return self._registrar.configs.get_member(_type, ensure_exists=True)

    # noinspection PyUnresolvedReferences, PyProtectedMember
    def load(
        self,
        configs: Optional[Configurations] = None,
        configs_file: Optional[str] = None,
        configs_dir: Optional[str] = None,
        configure: bool = False,
        **kwargs: Any,
    ) -> Sequence[Registrator]:
        _class = self._get_registator_class()

        defaults = self._load_registrator_defaults(**kwargs)
        if configs is None:
            configs = self._load_registrators_configs()
        if configs_file is None:
            configs_file = configs.name
        if configs_dir is None:
            configs_dir = configs.dirs.conf.joinpath(f"{configs.key}.d")
        return self._load(
            self._registrar,
            configs=configs,
            configs_file=configs_file,
            configs_dir=configs_dir,
            configure=configure,
            includes=_class.INCLUDES,
            defaults=defaults,
        )



# -*- coding: utf-8 -*-
"""
lories.connectors.modbus.register
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from pymodbus.client import ModbusBaseClient

from lories.core.configs import ConfigurationError
from lories.typing import Resource
from lories.util import to_int

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal

FunctionType = Literal["holding_register", "input_register", "coil"]
DataType = ModbusBaseClient.DATATYPE


class ModbusRegister:
    address: int

    function: FunctionType
    type: DataType

    @classmethod
    def from_resource(cls, resource: Resource) -> ModbusRegister:
        address = to_int(resource.get("address"))
        function = resource.get("function", default="holding_register")
        return ModbusRegister(address, function, cls._build_type(resource))

    @staticmethod
    def _build_type(resource: Resource) -> DataType:
        _type = resource.get("data_type", default=None)
        if _type is not None:
            _type = _type.lower()
        else:
            _type = resource.type
        if isinstance(_type, str):
            if _type in ["bits"]:
                return DataType.BITS
            if _type in ["uint16", "uint"]:
                return DataType.UINT16
            if _type == "uint32":
                return DataType.UINT32
            if _type == "uint64":
                return DataType.UINT64
            if _type in ["int16", "int"]:
                return DataType.INT16
            if _type == "int32":
                return DataType.INT32
            if _type == "int64":
                return DataType.INT64
            if _type in ["float32", "float"]:
                return DataType.FLOAT32
            if _type == "float64":
                return DataType.FLOAT64
            if _type in ["str", "string"]:
                return DataType.STRING
        elif isinstance(_type, type):
            if issubclass(_type, int):
                return DataType.INT16
            if issubclass(_type, float):
                return DataType.FLOAT32
            if issubclass(_type, str):
                return DataType.STRING
        raise ConfigurationError(f"Invalid data type '{_type}'")

    # noinspection PyShadowingBuiltins
    def __init__(self, address: int, function: FunctionType, type: DataType) -> None:
        super().__init__()
        self.address = address

        if function not in ["holding_register", "input_register", "coil"]:
            raise ConfigurationError(f"Invalid register function '{function}'")
        self.function = function

        if not isinstance(type, DataType):
            raise ConfigurationError(f"Invalid register data type '{type}'")
        self.type = type

    @property
    def length(self) -> int:
        return self.type.value[1]



# -*- coding: utf-8 -*-
"""
lories.components.camera
~~~~~~~~~~~~~~~~~~~~~~~~

"""

from __future__ import annotations

from typing import TypeVar

from lories.components import Component, register_component_type
from lories.core import Configurations


# noinspection SpellCheckingInspection
@register_component_type("camera")
class Camera(Component):
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        self.data.add(
            key="frame",
            name="Frame",
            type=bytes,
            aggregation="last",
        )

    # def activate(self) -> None:
    #     super().activate()
    #     self.data.register(
    #         self._on_frame,
    #         self.data.frame,
    #         how="all",
    #         unique=False,
    #     )
    #
    # def _on_frame(self, data: pd.DataFrame) -> None:
    #     pass


CameraType = TypeVar("CameraType", bound=Camera)



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks.connect
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from lories._core._channel import ChannelState  # noqa
from lories._core._connector import Connector  # noqa
from lories.connectors.tasks.task import ConnectorTask


class ConnectTask(ConnectorTask):
    # noinspection PyProtectedMember
    def run(self) -> Connector:
        self.connector.set_channels(ChannelState.CONNECTING)
        self.connector.connect(self.channels)
        self.connector.set_channels(ChannelState.CONNECTED)

        return self.connector



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.view
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Sequence, Type, TypeVar

import dash_bootstrap_components as dbc
from dash import html

from lories.application.view.pages import Page, PageFooter, PageGroup, PageHeader, PageLayout, PageRegistry
from lories.components import Component, ComponentContext
from lories.system import System
from lories.util import validate_key

P = TypeVar("P", bound=Page)
C = TypeVar("C", bound=Component)

registry = PageRegistry()


# noinspection PyShadowingBuiltins
def register_component_page(
    *types: Type[C],
    factory: Optional[Callable] = None,
    replace: bool = False,
) -> Callable[[Type[P]], Type[P]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[P]) -> Type[P]:
        registry.register_page(cls, *types, factory=factory, replace=replace)
        return cls

    return _register


def register_component_group(
    *types: Type[C],
    key: Optional[str] = None,
    name: Optional[str] = None,
    factory: Optional[Callable] = None,
    custom: bool = False,
    replace: bool = False,
) -> Callable[[Type[P]], Type[P]]:
    if name is None:
        name = types[0].__name__
    if key is None:
        key = validate_key(name)

    # noinspection PyShadowingBuiltins
    def _register(cls: Type[P]) -> Type[P]:
        if custom:
            if not issubclass(cls, PageGroup):
                raise ValueError(f"Unknown page group type: {cls.__name__}")
            type = cls
        else:
            type = PageGroup

        registry.register_group(type, *types, key=key, name=name, factory=factory, replace=replace)
        return cls

    return _register


class View(PageGroup):
    groups: Dict[str, PageGroup]

    # noinspection PyShadowingBuiltins
    def __init__(self, id: str, header: PageHeader, footer: PageFooter, *args, **kwargs) -> None:
        super().__init__(id=f"{id}-view", key="view", name="View", *args, **kwargs)
        self.header = header
        self.footer = footer

        self.groups = {}

    @property
    def path(self) -> str:
        return "/"

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def create_pages(self, components: ComponentContext) -> None:
        for component in components.values():
            if isinstance(component, System) or not isinstance(component.context, Component):
                self._create_page(self, component)

    # noinspection PyUnresolvedReferences
    def _create_page(self, view: PageGroup, component: Component) -> Optional[Page]:
        if not component.is_enabled():
            self._logger.debug(f"Skipping page creation for disabled {type(component).__name__} '{component.id}'")
            return None

        _type = type(component)
        if not registry.has_page(_type):
            return None

        registration = registry.get_page(_type)
        page = registration.initialize(
            component=component,
            group=view,
        )
        if page is not None:
            group = self._get_group(component)
            if group is not None:
                group.append(page)

            if isinstance(page, PageGroup):
                for page_component in page.components.values():
                    self._create_page(page, page_component)

            view.append(page)
        return page

    def _create_group(self, component: Any) -> Optional[PageGroup]:
        _type = type(component)
        if not registry.has_group(_type):
            return None

        registration = registry.get_group(_type)
        group = registration.initialize(
            id=f"{self.id}-{registration.key}",
            key=registration.key,
            name=registration.name,
        )
        if group is not None:
            self.groups[registration.key] = group
        return group

    def _get_group(self, component: Any) -> Optional[PageGroup]:
        _type = type(component)
        if not registry.has_group(_type):
            return None
        group = self.groups.get(registry.get_group(_type).key, None)
        if group is None:
            group = self._create_group(component)
        return group

    # noinspection PyProtectedMember
    def create_layout(self, layout: PageLayout) -> None:
        label = "Systems"
        menu = [dbc.DropdownMenuItem("Select", header=True)]
        for page in self._pages:
            if page.key in ["auth", "login"]:
                continue
            if not isinstance(page._component, System):
                label = "Select"
            menu.append(dbc.DropdownMenuItem(page.name, href=page.path))

        layout.menu = dbc.DropdownMenu(children=menu, nav=True, in_navbar=True, label=label)
        layout.container.class_name = "card-container card-focus"
        layout.container.children = []
        for page in self._pages:
            page_cards = []
            if isinstance(page, PageGroup):
                layout.append(
                    dbc.Row(dbc.Col(html.A(html.H4(f"{page.name}:"), className="card-header", href=page.path)))
                )
                for page_group in page:
                    if page_group.layout.has_card_items():
                        page_cards.append(page_group.layout.card)
            else:
                if page.layout.has_card_items():
                    page_cards.append(page.layout.card)

            layout.append(dbc.Row([dbc.Col(card, width="auto") for card in page_cards]))

    # noinspection PyTypeChecker, PyUnresolvedReferences
    def _at_create_layout(self, layout: PageLayout) -> None:
        super()._at_create_layout(layout)
        for page in self:
            page.create_layout(page.layout)
        for group in self.groups.values():
            group.create_layout(group.layout)
            if group.layout.has_menu_item():
                self.header.menu.append(group.layout.menu)

    def _on_create_layout(self, layout: PageLayout) -> None:
        super()._on_create_layout(layout)
        self.header.menu.insert(0, layout.menu)

    def _at_register(self) -> None:
        super()._at_register()
        groups = self.groups.values()

        def _register(pages: Sequence[Page]) -> None:
            for page in pages:
                if page.is_created() and page not in groups:
                    page.register()
                if isinstance(page, PageGroup):
                    _register(page)

        _register(self)
        for group in groups:
            group.register()
